# Test fixture for numbers
#
# The numbers module isn't properly supported, but we want to test that mypy
# can tell that it doesn't work as expected.

class Number: pass
class Complex: pass
class Real: pass
class Rational: pass
class Integral: pass
